import logging

from django.conf import settings
from django.core.cache import caches
import requests


_cache_key = '-'.join
logger = logging.getLogger(__name__)


def _fetch(suffix, params=None):
    """Return a JSON result from either a web API. If not configured
    correctly, throw a warning, but proceed"""
    if not settings.API_BASE:
        logger.error("API_BASE not configured. We won't have data")
        return None

    response = requests.get(settings.API_BASE + suffix, params=params)
    if response.status_code == requests.codes.ok:
        return response.json()
    elif response.status_code == 404:
        logger.warning("404 when fetching %s", settings.API_BASE + suffix)
        return None
    else:
        response.raise_for_status()


class ApiReader(object):
    """ Access the regulations API. Either hit the cache, or if there's a miss,
    hit the API instead and cache the results. """

    def __init__(self):
        self.cache = caches['api_cache']

    def all_regulations_versions(self):
        """ Get all versions, for all regulations. """
        return self._get('regulation')

    def regversions(self, label):
        return self._get('regulation/{}'.format(label))

    def cache_root_and_interps(self, reg_tree, version, is_root=True):
        """We will re-use the root tree at multiple points during page
        rendering, so cache it now. If caching an interpretation, also store
        child interpretations with titles (so that, when rendering slide-down
        interpretations, we don't perform additional fetches)"""
        if is_root or reg_tree.get('title'):
            tree_id = '-'.join(reg_tree['label'])
            cache_key = _cache_key(['regulation', tree_id, version])
            self.cache.set(cache_key, reg_tree)

        for child in reg_tree['children']:
            if child.get('node_type') == 'interp':
                self.cache_root_and_interps(child, version, False)

    def regulation(self, label, version):
        cache_key = _cache_key(['regulation', label, version])
        cached = self.cache.get(cache_key)

        if cached is not None:
            return cached
        else:
            regulation = _fetch('regulation/{}/{}'.format(label, version))
            # Add the tree to the cache
            if regulation:
                self.cache_root_and_interps(regulation, version)
                return regulation

    def _get(self, api_suffix, api_params=None):
        """ Retrieve from the cache whenever possible, or get from the API """
        if api_params is None:
            api_params = {}
        cache_key_elements = api_suffix.split('/') + list(sorted(
            element for pair in api_params.items() for element in pair))
        cache_key = _cache_key(cache_key_elements)
        cached = self.cache.get(cache_key)

        if cached is not None:
            return cached
        else:
            element = _fetch(api_suffix, api_params)
            self.cache.set(cache_key, element)
            return element

    def layer(self, layer_name, doc_type, label_id, version=None):
        """When retrieving layer data, we cheat a bit -- we always retrieve
        layer data corresponding to the "root" of the document, rather than
        only a subnode. We also must convert to the API format, where any
        version information is prefixed to doc_id"""
        root = label_id.split('-')[0]
        if version is None:
            doc_id = root
        else:
            doc_id = '{}/{}'.format(version, root)
        result = self._get(
            'layer/{}/{}/{}'.format(layer_name, doc_type, doc_id))
        # To remove - also try the old format for CFR layers; the API may not
        # have been updated
        if result is None and doc_type == 'cfr':
            result = self._get(
                'layer/{}/{}/{}'.format(layer_name, root, version))
        return result

    def diff(self, label, older, newer):
        """ End point for diffs. """
        return self._get("diff/{}/{}/{}".format(label, older, newer))

    def notices(self, part=None):
        """ End point for notice searching. Right now just a list. """
        params = None
        if part:
            params = {'part': part}
        return self._get('notice', params)

    def notice(self, fr_document_number):
        """ End point for retrieving a single notice. """
        return self._get('notice/{}'.format(fr_document_number))

    def search(self, query, doc_type='cfr', **kwargs):
        """Search via the API. Never cache these (that's the duty of the search
        index)"""
        kwargs['q'] = query
        return _fetch('search/{0}'.format(doc_type), kwargs)

    def preamble(self, doc_number):
        return self._get('preamble/{}'.format(doc_number))



from regulations.generator.layers.base import ParagraphLayer
from regulations.generator.layers.utils import convert_to_python


class MetaLayer(ParagraphLayer):
    shorthand = 'meta'
    data_source = 'meta'

    def __init__(self, layer_data):
        self.layer_data = convert_to_python(layer_data)

    def attach_metadata(self, node):
        """Return a pair of field-name (meta) + the layer data"""
        text_index = node['label_id']
        if self.layer_data.get(text_index):
            node['meta'] = self.layer_data[text_index][0]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-07-18 15:49
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('regulations', '0002_remove_failedcommentsubmission_files'),
    ]

    operations = [
        migrations.DeleteModel(
            name='FailedCommentSubmission',
        ),
    ]



from unittest import TestCase

from mock import patch

from regulations.generator import generator


class GeneratorTest(TestCase):
    @patch('regulations.generator.generator.api_reader')
    def test_get_tree_paragraph(self, api_reader):
        node = {'some': 'text'}
        api_reader.ApiReader.return_value.regulation.return_value = node

        p = generator.get_tree_paragraph('some-id', 'some-version')
        self.assertEqual(node, p)
        self.assertEqual(
            ('some-id', 'some-version'),
            api_reader.ApiReader.return_value.regulation.call_args[0])

    @patch('regulations.generator.generator.api_reader')
    def test_get_diff_json(self, api_reader):
        diff = {'some': 'diff'}
        api_reader.ApiReader.return_value.diff.return_value = diff
        d = generator.get_diff_json('204', 'old', 'new')
        self.assertEqual(diff, d)
        self.assertEqual(
            ('204', 'old', 'new'),
            api_reader.ApiReader.return_value.diff.call_args[0])

    @patch('regulations.generator.generator.api_reader')
    def test_get_notice(self, api_reader):
        notice = {'some': 'notice'}
        api_reader.ApiReader.return_value.notice.return_value = notice
        n = generator.get_notice('204-1234')
        self.assertEqual(notice, n)
        self.assertEqual(
            ('204-1234',),
            api_reader.ApiReader.return_value.notice.call_args[0])

    @patch('regulations.generator.generator.get_diff_json')
    def test_get_diff_applier(self, get_diff_json):
        diff = {'some': 'diff'}
        get_diff_json.return_value = diff
        da = generator.get_diff_applier('204', 'old', 'new')
        self.assertEqual(da.diff, diff)
        self.assertEqual(
            ('204', 'old', 'new'),
            get_diff_json.call_args[0])

    @patch('regulations.generator.generator.get_diff_json')
    def test_get_diff_applier_allows_empty(self, get_diff_json):
        diff = {}
        get_diff_json.return_value = diff
        da = generator.get_diff_applier('204', 'old', 'new')
        self.assertEqual(da.diff, diff)
        self.assertEqual(
            ('204', 'old', 'new'),
            get_diff_json.call_args[0])



from django.conf.urls import url

from notice_comment.views import (PrepareCommentView, SubmitCommentView,
                                  get_federal_agencies, get_gov_agency_types,
                                  preview_comment, upload_proxy)

urlpatterns = [
    url(r'^attachment$', upload_proxy),
    url(r'^review/(?P<doc_number>[\w-]+)$', PrepareCommentView.as_view(),
        name='comment_review'),
    url(r'^preview$', preview_comment),
    url(r'^(?P<doc_number>[\w-]+)$', SubmitCommentView.as_view(),
        name='comment_submit'),
    url(r'^federal_agencies$', get_federal_agencies),
    url(r'^gov_agency_types$', get_gov_agency_types),
]



from concurrent.futures import ThreadPoolExecutor
from importlib import import_module

from django.conf import settings

from regulations.generator import api_reader
from regulations.generator.layers.diff_applier import DiffApplier
from regulations.generator import notices


def _data_layers():
    """Index all configured data layers by their "shorthand". This doesn't
    have any error checking -- it'll explode if configured improperly"""
    layers = {}
    for class_path in settings.DATA_LAYERS:
        module, class_name = class_path.rsplit('.', 1)
        klass = getattr(import_module(module), class_name)
        layers[klass.shorthand] = klass
    return layers


DATA_LAYERS = _data_layers()


def generate_layers(layer_names, fetch_fn, **layer_attrs):
    """Return the three LayerApplier classes, populated with the appropriate
    layer data. Fetches this data in parallel.
    :param layer_names: list of layer short names
    :param fetch_fn: a function, which, when given a layer short name, returns
        the corresponding layer data
    :param layer_attrs: any other attributes to set on the layer object
    """
    layer_names = [l for l in layer_names if l in DATA_LAYERS]

    with ThreadPoolExecutor(max_workers=len(layer_names)) as executor:
        result_data = executor.map(fetch_fn, layer_names)

    for layer_name, layer_json in zip(layer_names, result_data):
        if layer_json is not None:
            layer_class = DATA_LAYERS[layer_name]
            layer = layer_class(layer_json)
            for attr_name, attr_val in layer_attrs.items():
                setattr(layer, attr_name, attr_val)

            yield layer


def layers(layer_names, doc_type, label_id, sectional=False, version=None):
    """Generate the three layer appliers for most situations"""
    def layer_fn(layer_name):
        api_layer_name = DATA_LAYERS[layer_name].data_source
        reader = api_reader.ApiReader()
        return reader.layer(api_layer_name, doc_type, label_id, version)
    return generate_layers(layer_names, layer_fn, version=version,
                           sectional=sectional)


def diff_layers(versions, label_id):
    """Generate the three layer appliers for diffs, which combine two sources
    of layer data"""
    def layer_fn(layer_name):
        api_layer_name = DATA_LAYERS[layer_name].data_source
        reader = api_reader.ApiReader()
        older_layer = reader.layer(api_layer_name, 'cfr', label_id,
                                   versions.older)
        newer_layer = reader.layer(api_layer_name, 'cfr', label_id,
                                   versions.newer)
        older_layer = older_layer or {}
        newer_layer = newer_layer or {}

        layer_json = dict(newer_layer)  # copy
        layer_json.update(older_layer)  # older layer takes precedence
        return layer_json
    layer_names = [
        'graphics', 'paragraph', 'keyterms', 'defined', 'formatting',
        'marker-hiding', 'marker-info'
    ]
    return generate_layers(layer_names, layer_fn, version=versions.older)


def get_tree_paragraph(paragraph_id, version):
    """Get a single level of the regulation tree."""
    api = api_reader.ApiReader()
    return api.regulation(paragraph_id, version)


def get_notice(document_number):
    """ Get a the data from a particular notice, given the Federal Register
    document number. """

    api = api_reader.ApiReader()
    return api.notice(document_number)


def get_sxs(label_id, notice, fr_page=None):
    """ Given a paragraph label_id, find the sxs analysis for that paragraph if
    it exists and has content. fr_page is used to distinguish between
    multiple analyses in the same notice."""

    all_sxs = notice['section_by_section']
    relevant_sxs = notices.find_label_in_sxs(all_sxs, label_id, fr_page)

    return relevant_sxs


def get_diff_json(regulation, older, newer):
    api = api_reader.ApiReader()
    return api.diff(regulation, older, newer)


def get_diff_applier(label_id, older, newer):
    regulation = label_id.split('-')[0]
    diff_json = get_diff_json(regulation, older, newer)
    if diff_json is not None:
        return DiffApplier(diff_json, label_id)



from django.template import loader

from regulations.generator.layers.base import InlineLayer


class DefinedLayer(InlineLayer):
    shorthand = 'defined'
    data_source = 'terms'

    def __init__(self, layer):
        self.layer = layer
        self.template = loader.get_template('regulations/layers/defining.html')

    def replacement_for(self, original, data):
        """Noop: never called"""
        pass

    def apply_layer(self, text, text_index):
        """Catch all terms which are defined in this paragraph, replace them
        with a span"""
        layer_pairs = []
        for ref_struct in self.layer['referenced'].values():
            if text_index == ref_struct['reference']:
                pos = tuple(ref_struct['position'])
                original = text[pos[0]:pos[1]]
                replacement = self.template.render({'term': original})
                replacement = replacement.strip('\n')
                layer_pairs.append((original, replacement, pos))
        return layer_pairs



import abc
from collections import namedtuple

from regulations.generator.layers.location_replace import LocationReplace


class LayerBase(object):
    """Base class for most layers; each layer contains information which is
    added on top of the regulation, such as definitions, internal citations,
    keyterms, etc."""
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def shorthand(self):
        """A short description for this layer. This is used in query strings
        and the like to define which layers should be used"""
        raise NotImplementedError

    @abc.abstractproperty
    def data_source(self):
        """Data is pulled from the API; this field indicates the name of the
        endpoint to pull data from"""
        raise NotImplementedError

    @abc.abstractmethod
    def inline_replacements(self, text_index, original_text):
        """Return triplets of (original text, replacement text, offsets)"""
        raise NotImplementedError

    @abc.abstractmethod
    def attach_metadata(self, node):
        """Attach metadata to the provided node"""
        raise NotImplementedError


Replacement = namedtuple('Replacement',
                         ['original', 'replacement', 'locations'])


class InlineLayer(LayerBase):
    """Represents a layer which replaces text by looking at offsets"""

    @abc.abstractmethod
    def replacement_for(self, original, data):
        """Given the original text and the relevant data from a layer, create
        a (string) replacement, by, for example, running the data through a
        template"""
        raise NotImplementedError

    def apply_layer(self, text, label_id):
        """Entry point when processing the regulation tree. Given the node's
        text and its label_id, yield all replacement text"""
        data_with_offsets = ((entry, start, end)
                             for entry in self.layer.get(label_id, [])
                             for (start, end) in entry['offsets'])
        for data, start, end in data_with_offsets:
            start, end = int(start), int(end)
            original = text[start:end]
            replacement = self.replacement_for(original, data)
            yield (original, replacement, (start, end))

    def inline_replacements(self, text_index, original_text):
        """Apply multiple inline layers to given text (e.g. links,
        highlighting, etc.)"""
        layer_pairs = self.apply_layer(original_text, text_index)

        # convert from offset-based to a search and replace layer.
        for original, replacement, offset in layer_pairs:
            offset_locations = LocationReplace.find_all_offsets(
                original, original_text)
            locations = [offset_locations.index(offset)]
            yield Replacement(original, replacement, locations)

    def attach_metadata(self, node):
        """Noop"""
        pass


class SearchReplaceLayer(LayerBase):
    """Represents a layer which replaces text by searching for and replacing a
    specific substring. Also accounts for the string appearing multiple times
    (via the 'locations' field)"""
    _text_field = 'text'    # All but key terms follow this convention...

    @abc.abstractmethod
    def replacements_for(self, text, data):
        """Given the original text and the relevant data from a layer, create
        a (string) replacement, by, for example, running the data through a
        template. Returns a generator"""
        raise NotImplementedError

    def inline_replacements(self, text_index, original_text):
        """Entry point when processing the regulation tree. Given the node's
        label_id, attempt to find relevant layer data in self.layer"""
        for entry in self.layer.get(text_index, []):
            text = entry[self._text_field]
            for replacement in self.replacements_for(text, entry):
                yield Replacement(text, replacement, entry['locations'])

    def attach_metadata(self, node):
        """Noop"""
        pass


class ParagraphLayer(LayerBase):
    """Represents a layer which applies meta data to nodes"""

    def inline_replacements(self, text_index, original_text):
        """Noop"""
        return []



import re

from six.moves.queue import PriorityQueue

from regulations.generator.layers.location_replace import LocationReplace


class LayersApplier(object):
    """ Most layers replace content. We try to do this intelligently here,
    so that layers don't step over each other. """
    HTML_TAG_REGEX = re.compile(r'<[^>]*?>')

    def __init__(self):
        self.queue = PriorityQueue()
        self.text = None

    def enqueue_from_list(self, elements_list):
        for le in elements_list:
            self.enqueue(le)

    def enqueue(self, layer_element):
        original, replacement, locations = layer_element
        priority = len(original)
        item = (original, replacement, locations)
        self.queue.put((-priority, item))

    def location_replace(self, xml_node, original, replacement, locations):
        LocationReplace().location_replace(xml_node, original, replacement,
                                           locations)

    def replace_all(self, original, replacement):
        """ Replace all occurrences of original with replacement. This is HTML
        aware; it effectively looks at all of the text in between HTML tags"""
        text_chunks = []
        index = 0
        for match in self.HTML_TAG_REGEX.finditer(self.text):
            text = self.text[index:match.start()]
            text_chunks.append(text.replace(original, replacement))
            text_chunks.append(self.text[match.start():match.end()])    # tag
            index = match.end()
        text_chunks.append(self.text[index:])   # trailing text
        self.text = "".join(text_chunks)

    def replace_at(self, original, replacement, locations):
        """ Replace the occurrences of original at all the locations with
        replacement. """

        locations.sort()
        self.text = LocationReplace().location_replace_text(
            self.text, original, replacement, locations)

    def apply_layers(self, original_text):
        self.text = original_text

        while not self.queue.empty():
            priority, layer_element = self.queue.get()
            original, replacement, locations = layer_element

            if not locations:
                self.replace_all(original, replacement)
            else:
                self.replace_at(original, replacement, locations)

        return self.text



from regulations.generator.sidebar.base import SidebarBase
from regulations.views.utils import regulation_meta


class PrintPart(SidebarBase):
    shorthand = 'print_part'

    def context(self, http_client, request):
        meta = regulation_meta(self.cfr_part, self.version)

        return {
            'cfr_title': meta.get('cfr_title_number'),
            'cfr_part': self.cfr_part,
            'version': self.version,
        }



from xml.etree import ElementTree as etree  # nosec - see usage below

from django.template import Context, Template


def _gen_link(content):
    """Shorthand for passing the content into a template and rendering"""
    text = "{% load macros %}" + content
    as_str = Template(text).render(Context({}))
    # Safe because: we've constructed the XML
    as_xml = etree.fromstring("<ROOT>{}</ROOT>".format(as_str))  # nosec
    anchors = as_xml.findall('.//a')
    assert len(anchors) > 0
    return anchors[0]


def test_external_link_no_optional():
    """The classes and title fields are optional. We should generate an
    appropriate link"""
    anchor = _gen_link(
        '{% external_link url="http://example.com/path" text="Click" %}')
    assert anchor.get('target') == '_blank'
    assert anchor.get('href') == 'http://example.com/path'
    assert 'title' not in anchor.attrib
    assert 'aria-label' in anchor.attrib
    assert 'Click' in anchor.text


def test_external_link_classes_title():
    """The classes and title fields _can_ be added"""
    anchor = _gen_link(
        '{% external_link url="url" text="text" classes="some here" '
        'title="My Title" %}')
    assert anchor.get('title') == 'My Title'
    assert 'some here' in anchor.get('class')


def test_external_link_is_escaped():
    # LXML decodes the value for us, so let's look at the raw markup
    text = ('{% load macros %}'
            '{% external_link url="http://example.com/?param=1&other=value" '
            'text="value" %}')
    as_str = Template(text).render(Context({}))
    assert 'http://example.com/?param=1&amp;other=value' in as_str


def test_search_for():
    """Macro should url-encode search terms."""
    anchor = _gen_link(
        '{% search_for terms="has spaces" reg="1234" version="vvv" %}')
    assert '1234' in anchor.get('href')
    assert 'vvv' in anchor.get('href')
    assert 'has%20spaces' in anchor.get('href')



from unittest import TestCase

from mock import patch

from regulations.generator.layers.paragraph_markers import (
    MarkerHidingLayer, MarkerInfoLayer, ParagraphMarkersLayer)


class ParagraphMarkersLayerTest(TestCase):
    @patch('regulations.generator.layers.paragraph_markers.loader')
    def test_replacements_for(self, loader):
        pml = ParagraphMarkersLayer({})
        result = list(pml.replacements_for('(a)', {}))
        self.assertEqual(1, len(result))
        call_args = loader.get_template.return_value.render.call_args[0][0]
        self.assertEqual('(a)', call_args['paragraph'])
        self.assertEqual('a', call_args['paragraph_stripped'])

        result = list(pml.replacements_for('q.', {}))
        self.assertEqual(1, len(result))
        call_args = loader.get_template.return_value.render.call_args[0][0]
        self.assertEqual('q.', call_args['paragraph'])
        self.assertEqual('q', call_args['paragraph_stripped'])


class MarkerHidingLayerTest(TestCase):
    @patch('regulations.generator.layers.paragraph_markers.loader')
    def test_replacements_for(self, loader):
        mhl = MarkerHidingLayer({})
        result = list(mhl.replacements_for('(a)', {}))
        self.assertEqual(1, len(result))
        call_args = loader.get_template.return_value.render.call_args[0][0]
        self.assertEqual('(a)', call_args['paragraph'])

        result = list(mhl.replacements_for('q.', {}))
        self.assertEqual(1, len(result))
        call_args = loader.get_template.return_value.render.call_args[0][0]
        self.assertEqual('q.', call_args['paragraph'])


class MarkerInfoLayerTest(TestCase):
    def test_attach_metadata(self):
        mil = MarkerInfoLayer({
            '1001-12-a': [{'text': '(a)', 'locations': [0]}],
            '1001-12-q': [{'text': 'q.', 'locations': [1]}]
        })
        node = {'label_id': '1002-01-01'}
        mil.attach_metadata(node)
        self.assertEqual(set(node.keys()), {'label_id'})

        node['label_id'] = '1001-12-a'
        mil.attach_metadata(node)
        self.assertEqual(node['paragraph_marker'], 'a')

        node['label_id'] = '1001-12-q'
        mil.attach_metadata(node)
        self.assertEqual(node['paragraph_marker'], 'q')



from django.http import HttpResponse
from django.template.loader import select_template


def about(request):
    t = select_template([
        'regulations/custom-about.html',
        'regulations/about.html'])
    return HttpResponse(t.render({}, request))



from mock import Mock

from regulations.generator.sidebar import print_part


def test_print_part(monkeypatch):
    monkeypatch.setattr(print_part, 'regulation_meta', Mock())
    print_part.regulation_meta.return_value = {'cfr_title_number': 12}

    sidebar = print_part.PrintPart('333-44', 'vvvv')
    assert sidebar.context(Mock(), Mock()) == {
        'cfr_title': 12,
        'cfr_part': '333',
        'version': 'vvvv',
    }



from setuptools import find_packages, setup


setup(
    name="regulations",
    version="8.4.2",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'cached-property',
        'django>=1.8,<1.12',
        'enum34;python_version<"3.4"',
        'futures;python_version<"3"',
        'requests',
        'six',
    ],
    extras_require={
        'notice_comment': ['boto3', 'celery', 'requests-toolbelt'],
    },
    classifiers=[
        'License :: Public Domain',
        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication'
    ]
)



from datetime import date

import pytest
from django.http import HttpResponseGone
from mock import call, Mock

from regulations.generator.versions import Timeline
from regulations.views import chrome, error_handling


def test_version_span():
    history = [
        {'by_date': date(2009, 9, 9), 'timeline': Timeline.future},
        {'by_date': date(2008, 8, 8), 'timeline': Timeline.present},
        {'by_date': date(2007, 7, 7), 'timeline': Timeline.past},
    ]
    result = chrome.version_span(history, date(2010, 10, 10))
    assert result == chrome.VersionSpan(date(2009, 9, 9), None,
                                        Timeline.future)

    result = chrome.version_span(history, date(2009, 9, 8))
    assert result == chrome.VersionSpan(date(2008, 8, 8), date(2009, 9, 9),
                                        Timeline.present)

    result = chrome.version_span(history, date(2007, 8, 9))
    assert result == chrome.VersionSpan(date(2007, 7, 7), date(2008, 8, 8),
                                        Timeline.past)


def test_chrome_404(monkeypatch, rf):
    """Test that the response of the outer view is that of the inner when
    there's an error"""
    monkeypatch.setattr(chrome.ChromeView, 'set_chrome_context', Mock())
    monkeypatch.setattr(chrome, 'generator', Mock())
    monkeypatch.setattr(error_handling, 'api_reader', Mock())

    regversions = error_handling.api_reader.ApiReader.return_value.regversions
    regversions.return_value = None
    chrome.generator.get_tree_paragraph.return_value = {}
    chrome.ChromeView.set_chrome_context.return_value = None

    class InnerView(chrome.TemplateView):
        def get(self, request, *args, **kwargs):
            return HttpResponseGone()

    class FakeView(chrome.ChromeView):
        partial_class = InnerView

    view = FakeView()
    view.request = rf.get('/')
    response = view.get(view.request, label_id='lab', version='ver')
    assert response.status_code == 404


def test_chrome_empty_meta(monkeypatch, rf):
    """Return 404 when trying to access missing regulation.
    In this situation, `regulation_meta` returns {} """

    chrome_view = chrome.ChromeView()

    monkeypatch.setattr(chrome, 'fetch_grouped_history', Mock())
    monkeypatch.setattr(chrome, 'fetch_toc', Mock(return_value=[]))
    monkeypatch.setattr(chrome.utils, 'regulation_meta', Mock(return_value={}))

    with pytest.raises(error_handling.MissingContentException):
        chrome_view.set_chrome_context({}, '2', 'version')


def test_chrome_error_propagation(monkeypatch, rf):
    """While we don't rely on this sort of propagation for the main content
    (much), test it in the sidebar"""
    monkeypatch.setattr(chrome, 'error_handling', Mock())
    monkeypatch.setattr(chrome, 'generator', Mock())
    monkeypatch.setattr(chrome, 'SideBarView', Mock())
    chrome.SideBarView.as_view.return_value.return_value = HttpResponseGone()

    class FakeView(chrome.ChromeView):
        def add_main_content(self, context):
            pass

        def set_chrome_context(self, context, reg_part, version):
            pass

    view = FakeView()
    view.request = rf.get('/')
    response = view.get(view.request, label_id='lab', version='ver')
    assert response.status_code == 410


def test_chrome_get_404(client, monkeypatch):
    monkeypatch.setattr(chrome, 'generator', Mock())
    chrome.generator.get_regulation.return_value = None
    response = client.get('/regulation/111/222')
    assert response.status_code == 404


def test_chrome_get_404_tree(client, monkeypatch):
    monkeypatch.setattr(chrome, 'generator', Mock())
    chrome.generator.get_regulation.return_value = {'regulation': 'tree'}
    chrome.generator.get_tree_paragraph.return_value = None
    response = client.get('/regulation/111/222')
    assert response.status_code == 404


def test_chrome_diff_redirect_label_regulation():
    """If viewing a full regulation, the redirect for diffs should point to
    the first section"""
    view = chrome.ChromeView()
    toc = [{'section_id': '199-Subpart-A',
            'sub_toc': [{'section_id': '199-4'}, {'section_id': '199-6'}]},
           {'section_id': '199-Subpart-B',
            'sub_toc': [{'section_id': '199-8'}, {'section_id': '199-9'}]}]
    assert view.diff_redirect_label('199', toc) == '199-4'


@pytest.mark.parametrize('label, expected', (
    ('199-4-b', '199-4'),
    ('199-4-b-3', '199-4'),
    ('199-A', '199-A'),
    ('199-A-14B', '199-A'),
    ('199-Interp', '199-Interp'),
    ('199-Interp-h1', '199-Interp'),
    ('199-12-Interp-2', '199-Interp'),
))
def test_chrome_diff_redirect_label_paragraph(label, expected):
    """If viewing a single paragraph, the redirect for diffs should point to
    that paragraph's section. Similarly, all diffs for interpretations should
    point to the root interpretation"""
    view = chrome.ChromeView()
    assert view.diff_redirect_label(label, []) == expected


@pytest.mark.parametrize('label', (
    '199-Subpart-Interp', '199-Subpart-A-Interp', '199-Appendices-Interp'))
def test_subterp_diff_redirect_label(label):
    view = chrome.ChromeSubterpView()
    assert '199-Interp' == view.diff_redirect_label(label, None)


def test_subterp_check_tree(monkeypatch):
    monkeypatch.setattr(chrome, 'generator', Mock())
    monkeypatch.setattr(chrome, 'filter_by_subterp', Mock())
    chrome.generator.get_tree_paragraph.return_value = None

    view = chrome.ChromeSubterpView()
    with pytest.raises(chrome.error_handling.MissingSectionException):
        view.check_tree({'version': 'vvvv', 'label_id': 'llll'})

    chrome.generator.get_tree_paragraph.return_value = {'children': []}
    chrome.filter_by_subterp.return_value = []
    with pytest.raises(chrome.error_handling.MissingSectionException):
        view.check_tree({'version': 'vvvv', 'label_id': 'llll'})

    chrome.filter_by_subterp.return_value = ["something"]
    # No exception
    view.check_tree({'version': 'vvvv', 'label_id': 'llll'})


def test_chrome_search_version_present(monkeypatch, rf):
    """If a version is in the request, we use it to derive the label_id."""
    monkeypatch.setattr(chrome, 'utils', Mock())
    chrome.utils.first_section.return_value = '111-22'

    view = chrome.ChromeSearchView()
    view.request = rf.get('/?version=some-version')
    result = view.fill_kwargs({'label_id': '111'})

    assert result == {
        'version': 'some-version',
        'skip_count': True,
        'label_id': '111-22',
    }
    assert chrome.utils.first_section.call_args == call('111', 'some-version')


def test_chome_search_missing_version(monkeypatch, rf):
    """A missing version shouldn't cause the results page to explode"""
    monkeypatch.setattr(chrome, 'utils', Mock())
    monkeypatch.setattr(chrome, 'get_versions', Mock())
    chrome.get_versions.return_value = (
        {'version': 'current-version'}, {'version': 'next-version'})
    chrome.utils.first_section.return_value = '222-33'

    view = chrome.ChromeSearchView()
    view.request = rf.get('/')
    result = view.fill_kwargs({'label_id': '222'})

    assert result == {
        'version': 'current-version',
        'skip_count': True,
        'label_id': '222-33',
    }
    assert chrome.get_versions.call_args == call('222')
    assert chrome.utils.first_section.call_args == call('222',
                                                        'current-version')



from .base import *

DEBUG = True

CACHES['default']['BACKEND'] = 'django.core.cache.backends.dummy.DummyCache'
CACHES['eregs_longterm_cache']['BACKEND'] = 'django.core.cache.backends.dummy.DummyCache'
CACHES['api_cache']['TIMEOUT'] = 5  # roughly per request

ROOT_URLCONF = 'regulations.all_urls'
INSTALLED_APPS = INSTALLED_APPS + ('notice_comment',)

try:
    from local_settings import *
except ImportError:
    pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from datetime import date

from mock import Mock

from regulations.generator import section_url
from regulations.generator.versions import Timeline
from regulations.views import partial_search


def test_get(client, monkeypatch):
    monkeypatch.setattr(partial_search, 'api_reader', Mock())
    monkeypatch.setattr(partial_search, 'fetch_grouped_history', Mock())
    partial_search.api_reader.ApiReader.return_value.search.return_value = {
        'total_hits': 3333,
        'results': [
            {'label': ['111', '22'], 'text': 'tttt', 'version': 'vvv',
             'title':"Consumer's"},
            {'label': ['111', '24', 'a'], 'text': 'o', 'version': 'vvv'},
            {'label': ['111', '25'], 'text': 'more', 'version': 'vvv'}
        ]
    }
    partial_search.fetch_grouped_history.return_value = [
        {'notices': [{'document_number': 'bbb',
                      'effective_on': date(2012, 12, 12)}]},
        {'notices': [{'document_number': 'ccc',
                      'effective_on': date(2001, 1, 1)},
                     {'document_number': 'vvv',
                      'effective_on': date(2003, 4, 5)}],
         'timeline': Timeline.past}
    ]
    response = client.get('/partial/search/111?version=vvv&q=none').content
    assert b'111-22' in response
    assert b'111-24-a' in response
    assert b'111.24(a)' in response
    assert b'111-25' in response
    assert b'111.25' in response
    assert b'3333' in response
    assert b'Consumer&#39;s' in response


def test_root_info(client, monkeypatch):
    monkeypatch.setattr(partial_search, 'api_reader', Mock())
    monkeypatch.setattr(partial_search, 'fetch_grouped_history', Mock())
    partial_search.api_reader.ApiReader.return_value.search.return_value = {
        'total_hits': 2,
        'results': [
            {'label': ['444', '22'], 'text': 'tttt', 'version': 'vvv',
             'title':"consumer's"},
            {'label': ['444', '24', 'a'], 'text': 'o', 'version': 'vvv'},
        ]
    }
    partial_search.fetch_grouped_history.return_value = [
        {'notices': [{'document_number': 'bbb',
                      'effective_on': date(2012, 12, 12)}]},
        {'notices': [{'document_number': 'ccc',
                      'effective_on': date(2001, 1, 1)},
                     {'document_number': 'vvv',
                      'effective_on': date(2003, 4, 5)}],
         'timeline': Timeline.past}
    ]
    response = client.get('/partial/search/444?version=vvv&q=none').content
    assert b'2 results' in response


def test_subinterp(client, monkeypatch):
    monkeypatch.setattr(partial_search, 'api_reader', Mock())
    monkeypatch.setattr(partial_search, 'fetch_grouped_history', Mock())
    monkeypatch.setattr(section_url, 'fetch_toc', Mock())

    section_url.fetch_toc.return_value = [
        {'index': ['444', 'Subpart', 'B'], 'is_subpart': True,
         'section_id': '444-Subpart-B', 'sub_toc': [
            {'index': ['444', '22'], 'section_id': '444-22',
             'is_section': True}]},
        {'index': ['444', 'Interp'], 'section_id': '444-Interp',
         'sub_toc': [
            {'index': ['444', 'Interp', 'h1'],
             'section_id': '444-Interp-h1'},
            {'index': ['444', 'Subpart', 'B', 'Interp'],
             'section_id': '444-Subpart-B-Interp'}]}
    ]
    partial_search.api_reader.ApiReader.return_value.search.return_value = {
        'total_hits': 3,
        'results': [
            {'label': ['444', '22', 'Interp'], 'text': 'tttt',
             'version': 'vvv', 'title':"consumer's"},
            {'label': ['444', 'Interp', 'h1', 'p5'], 'text': 'o',
             'version': 'vvv'}
        ]
    }
    partial_search.fetch_grouped_history.return_value = [
        {'notices': [{'document_number': 'bbb',
                      'effective_on': date(2012, 12, 12)}]},
        {'notices': [{'document_number': 'ccc',
                      'effective_on': date(2001, 1, 1)},
                     {'document_number': 'vvv',
                      'effective_on': date(2003, 4, 5)}],
         'timeline': Timeline.past}
    ]

    response = client.get('/partial/search/444?version=vvv&q=other').content
    assert b'444-Subpart-B-Interp' in response
    assert b'444-Interp-h1' in response


def test_no_results(client, monkeypatch):
    monkeypatch.setattr(partial_search, 'api_reader', Mock())
    monkeypatch.setattr(partial_search, 'fetch_grouped_history', Mock())
    partial_search.api_reader.ApiReader.return_value.search.return_value = {
        'total_hits': 0,
        'results': []
    }
    partial_search.fetch_grouped_history.return_value = [
        {'notices': [{'document_number': 'bbb',
                      'effective_on': date(2012, 12, 12)}]},
        {'notices': [{'document_number': 'ccc',
                      'effective_on': date(2001, 1, 1)},
                     {'document_number': 'vvv',
                      'effective_on': date(2003, 4, 5)}],
         'timeline': Timeline.past}
    ]
    response = client.get('/partial/search/121?version=vvv&q=none').content
    assert b'4/5/2003' in response


def test_null_params(client, monkeypatch):
    monkeypatch.setattr(partial_search, 'fetch_grouped_history',
                        Mock(return_value=[]))

    response = client.get('/partial/search/111?version=vvv').content
    assert b'provide a query' in response

    response = client.get('/partial/search/111?q=vvv').content
    assert b'provide a version' in response


def test_preamble_search(client, monkeypatch):
    monkeypatch.setattr(partial_search, 'api_reader', Mock())
    partial_search.api_reader.ApiReader.return_value.search.return_value = {
        'total_hits': 3333,
        'results': [
            {'label': ['111_22', 'I', 'A'], 'text': 'tttt',
             'title': 'A. Something'},
            {'label': ['111_22', 'I', 'p1'], 'text': 'eee'}
        ]
    }

    response = client.get('/partial/search/preamble/111?q=none').content
    assert b'111_22-I-A' in response
    assert b'111_22-I-p1' in response
    assert b'tttt' in response
    assert b'eee' in response
    assert b'A. Something' in response
    assert b'Section I.A' in response


def test_page_size(client, monkeypatch):
    """Page size should be passed along to the API"""
    monkeypatch.setattr(partial_search, 'api_reader', Mock())
    search = partial_search.api_reader.ApiReader.return_value.search
    search.return_value = {'total_hits': 0, 'results': []}

    client.get('/partial/search/preamble/111?q=none')

    assert search.call_args[1]['page_size'] == 10


def test_add_prev_next():
    view = partial_search.PartialSearch()
    context = {'results': {'total_hits': 77}}
    view.add_prev_next(0, context)
    assert 'previous' not in context
    assert context['next'] == {'page': 1, 'length': 10}

    del context['next']
    view.add_prev_next(5, context)
    assert context['previous'] == {'page': 4, 'length': 10}
    assert context['next'] == {'page': 6, 'length': 10}

    del context['previous']
    del context['next']
    view.add_prev_next(6, context)
    assert context['previous'] == {'page': 5, 'length': 10}
    assert context['next'] == {'page': 7, 'length': 7}

    del context['previous']
    del context['next']
    view.add_prev_next(7, context)
    assert context['previous'] == {'page': 6, 'length': 10}
    assert 'next' not in context


def test_add_cfr_headers_no_title():
    """If no title is present in the result data, we construct one based on
    its label."""
    result = partial_search.add_cfr_headers({'label': ['111', '22', 'c', '4']})
    assert result['header'] == '111.22(c)(4)'


def test_add_cfr_headers_section():
    """If all of the titles are the same (as is the case when a section's
    intro text matches our search), we don't need to repeat them."""
    result = partial_search.add_cfr_headers({'title': 'Section 22',
                                             'match_title': 'Section 22',
                                             'paragraph_title': 'Section 22'})
    assert result['header'] == 'Section 22'
    assert 'subheader' not in result
    assert 'subsubheader' not in result


def test_add_cfr_headers_three_levels():
    """If the three levels of title are different, we expect them in our
    results."""
    result = partial_search.add_cfr_headers({
        'title': 'Section 22',
        'match_title': 'A matching heading',
        'paragraph_title': 'A title for text',
    })
    assert result['header'] == 'Section 22'
    assert result['subheader'] == 'A matching heading'
    assert result['subsubheader'] == 'A title for text'
